# pyshimmer - API for Shimmer sensor devices
# Copyright (C) 2020  Lukas Magel

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import struct
from typing import Tuple

from serial import Serial

from pyshimmer.dev.base import sec2ticks, ticks2sec
from pyshimmer.dev.exg import ExGRegister
from pyshimmer.dev.fw_version import get_firmware_type, EFirmwareType
from pyshimmer.uart.dock_const import *
from pyshimmer.uart.dock_serial import DockSerial
from pyshimmer.util import unpack


class ShimmerDock:
    """Main API to communicate with the Shimmer over the Dock UART

    :arg ser: The serial interface to use for communication

    """

    def __init__(self, ser: Serial, flush_before_req=True):
        self._serial = DockSerial(ser)
        self._flush_before_req = flush_before_req

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.close()

    def _read_resp_type_or_throw(self, expected: int) -> int:
        r = self._serial.read_byte()
        if r != START_CHAR:
            raise IOError(f'Unknown start character encountered: {r:x}')

        cmd = self._serial.read_byte()
        if cmd == UART_BAD_ARG_RESPONSE:
            raise IOError('Command failed: Bad argument')
        elif cmd == UART_BAD_CMD_RESPONSE:
            raise IOError('Command failed: Unknown command')
        elif cmd == UART_BAD_CRC_RESPONSE:
            raise IOError('Command failed: CRC Error')
        elif cmd != expected:
            raise IOError(f'Wrong response type: {expected:x} != {cmd:x}')

        return cmd

    def _write_packet(self, cmd: int, comp: int, prop: int, data: bytes = bytes()) -> None:
        if self._flush_before_req:
            self._serial.flush_input_buffer()
        self._serial.start_write_crc()

        pkt_len = 2 + len(data)
        self._serial.write_packed('<BBBBB', START_CHAR, cmd, pkt_len, comp, prop)
        self._serial.write(data)

        self._serial.end_write_crc()

    def _write_packet_wformat(self, cmd: int, comp: int, prop: int, fmt: str, *args: any) -> None:
        data = struct.pack(fmt, *args)
        self._write_packet(cmd, comp, prop, data)

    def _read_response(self) -> Tuple[int, int, bytes]:
        self._serial.start_read_crc_verify()

        self._read_resp_type_or_throw(UART_RESPONSE)
        pkt_len, comp, prop = self._serial.read_packed('BBB')

        data_len = pkt_len - 2
        data = self._serial.read(data_len)

        self._serial.end_read_crc_verify()
        return comp, prop, data

    def _read_response_verify(self, exp_comp: int, exp_prop: int) -> bytes:
        comp, prop, data = self._read_response()

        if exp_comp != comp:
            raise IOError(f'Encountered unexpected component type in response: {exp_comp:x} != {comp:x}')
        elif exp_prop != prop:
            raise IOError(f'Encountered unexpected property type in response: {exp_prop:x} != {prop:x}')

        return data

    def _read_response_wformat_verify(self, exp_comp: int, exp_prop: int, fmt: str) -> any:
        data_packed = self._read_response_verify(exp_comp, exp_prop)
        data = struct.unpack(fmt, data_packed)
        return unpack(data)

    def _read_ack(self) -> None:
        self._serial.start_read_crc_verify()
        self._read_resp_type_or_throw(UART_ACK_RESPONSE)
        self._serial.end_read_crc_verify()

    def close(self) -> None:
        """Close the underlying serial interface and release all resources

        """
        self._serial.close()

    def get_mac_address(self) -> Tuple[int, ...]:
        """Retrieve the Bluetooth MAC address of the device

        :return: A tuple containing six integer values, each representing a single byte of the address
        """
        self._write_packet(UART_GET, UART_COMP_SHIMMER, UART_PROP_MAC)

        mac = self._read_response_wformat_verify(UART_COMP_SHIMMER, UART_PROP_MAC, 'BBBBBB')
        return mac

    def set_rtc(self, ts_sec: float) -> None:
        """Set the real-time clock of the device

        Specify the UNIX timestamp in seconds as new value for the real-time clock of the device

        :param ts_sec: The UNIX timestamp in seconds
        """
        ticks = sec2ticks(ts_sec)
        self._write_packet_wformat(UART_SET, UART_COMP_SHIMMER, UART_PROP_RWC_CFG_TIME, '<Q', ticks)
        self._read_ack()

    def get_rtc(self) -> float:
        """Retrieve the current value of the real-time clock

        :return: A floating-point value representing the current value of the real-time clock as UNIX timestamp
            in seconds
        """
        self._write_packet(UART_GET, UART_COMP_SHIMMER, UART_PROP_CURR_LOCAL_TIME)
        ticks = self._read_response_wformat_verify(UART_COMP_SHIMMER, UART_PROP_CURR_LOCAL_TIME, '<Q')
        return ticks2sec(ticks)

    def get_config_rtc(self) -> float:
        """Get the value that was last set for the real-time clock

        Example:

            The real-time clock is set to a value of 42s. Subsequent calls to :meth:`get_rtc` will return v > 42s,
            while :meth:`get_config_rtc` will return 42s.

        :return: A floating-point value representing the last configured value for the real-time clock as UNIX
            timestamp in seconds
        """
        self._write_packet(UART_GET, UART_COMP_SHIMMER, UART_PROP_RWC_CFG_TIME)
        ticks = self._read_response_wformat_verify(UART_COMP_SHIMMER, UART_PROP_RWC_CFG_TIME, '<Q')
        return ticks2sec(ticks)

    def get_firmware_version(self) -> Tuple[int, EFirmwareType, int, int, int]:
        """Retrieve the firmware version of the device

        :return: A tuple containing the following values:
            - The hardware version, should be 3 for Shimmer3
            - The firmware type: LogAndStream or SDLog
            - The major release version
            - The minor release version
            - The patch level
        """
        self._write_packet(UART_GET, UART_COMP_SHIMMER, UART_PROP_VER)
        hw_ver, fw_type_bin, major, minor, rel = self._read_response_wformat_verify(UART_COMP_SHIMMER,
                                                                                    UART_PROP_VER, '<BHHBB')
        fw_type = get_firmware_type(fw_type_bin)
        return hw_ver, fw_type, major, minor, rel

    def get_firmware_type(self) -> EFirmwareType:
        """Retrieve the active firmware type

        :return: The firmware type: LogAndStream or SDLog
        """
        _, fw_type, _, _, _ = self.get_firmware_version()
        return fw_type

    def get_infomem(self, addr: int, dlen: int) -> bytes:
        """Access the infomem memory and retrieve the specified range

        :param addr: The start address
        :param dlen: The length of the memory block that will be retrieved
        :return: The bytes of the memory block
        """
        # Due to a bug in the firmware code, we must manually set a variable in the firmware to a specific value
        # using a different command before we can read the InfoMem.
        self._write_packet_wformat(UART_GET, UART_COMP_DAUGHTER_CARD, UART_PROP_CARD_ID, '<BH', 0x0, 0x0)
        self._read_response()

        self._write_packet_wformat(UART_GET, UART_COMP_SHIMMER, UART_PROP_INFOMEM, '<BH', dlen, addr)
        return self._read_response_verify(UART_COMP_SHIMMER, UART_PROP_INFOMEM)

    def get_exg_register(self, chip_id: int) -> ExGRegister:
        if not 0 <= chip_id <= 1:
            raise ValueError('Parameter chip_id must be 0 or 1')

        offset = UART_INFOMEM_EXG_OFFSET + chip_id * 0x0A
        dlen = 0x0A

        reg_data = self.get_infomem(offset, dlen)
        return ExGRegister(reg_data)



# pyshimmer - API for Shimmer sensor devices
# Copyright (C) 2020  Lukas Magel

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import struct
from typing import List, Tuple, Union, BinaryIO

import numpy as np

from pyshimmer.dev.channels import ESensorGroup, get_ch_dtypes, get_enabled_channels, EChannelType, \
    ENABLED_SENSORS_LEN, deserialize_sensors
from pyshimmer.dev.exg import ExGRegister
from pyshimmer.util import FileIOBase, unpack, bit_is_set
from .reader_const import RTC_CLOCK_DIFF_OFFSET, ENABLED_SENSORS_OFFSET, SR_OFFSET, \
    START_TS_OFFSET, START_TS_LEN, TRIAL_CONFIG_OFFSET, TRIAL_CONFIG_MASTER, TRIAL_CONFIG_SYNC, BLOCK_LEN, \
    DATA_LOG_OFFSET, EXG_REG_OFFSET, EXG_REG_LEN, TRIAXCAL_FILE_OFFSET, TRIAXCAL_OFFSET_SCALING, \
    TRIAXCAL_GAIN_SCALING, TRIAXCAL_ALIGNMENT_SCALING


class ShimmerBinaryReader(FileIOBase):

    def __init__(self, fp: BinaryIO):
        super().__init__(fp)

        self._sensors = []
        self._channels = []
        self._sr = 0
        self._rtc_diff = 0
        self._start_ts = 0
        self._trial_config = 0

        self._read_header()

    @staticmethod
    def get_data_channels(sensors):
        channels = get_enabled_channels(sensors)
        channels_with_ts = [EChannelType.TIMESTAMP] + channels
        return channels_with_ts

    def _read_header(self) -> None:
        self._sr = self._read_sample_rate()
        self._sensors = self._read_enabled_sensors()
        self._channels = self.get_data_channels(self._sensors)
        self._channel_dtypes = get_ch_dtypes(self._channels)
        self._rtc_diff = self._read_rtc_clock_diff()
        self._start_ts = self._read_start_time()
        self._trial_config = self._read_trial_config()
        self._exg_regs = self._read_exg_regs()

        self._samples_per_block, self._block_size = self._calculate_block_size()

    def _read_sample_rate(self) -> int:
        self._seek(SR_OFFSET)
        return self._read_packed('<H')

    def _read_enabled_sensors(self) -> List[ESensorGroup]:
        self._seek(ENABLED_SENSORS_OFFSET)
        sensor_bitfield = self._read(ENABLED_SENSORS_LEN)
        enabled_sensors = deserialize_sensors(sensor_bitfield)

        return enabled_sensors

    def _read_rtc_clock_diff(self) -> int:
        self._seek(RTC_CLOCK_DIFF_OFFSET)
        rtc_diff_ticks = self._read_packed('>Q')
        return rtc_diff_ticks

    def _read_start_time(self) -> int:
        self._seek(START_TS_OFFSET)
        ts_bin = self._read(START_TS_LEN)

        # The timestamp is 5 byte long in little endian byte order, but has its MSB at offset 0 instead of 4.
        # Due to this, we need to move the last byte back to the end, pad it to 8 bytes and parse it as 64bit value.
        ts_bin_flipped = ts_bin[1:] + ts_bin[0:1]
        ts_bin_padded = ts_bin_flipped + b'\x00' * 3

        ts_ticks = struct.unpack('<Q', ts_bin_padded)
        return unpack(ts_ticks)

    def _read_trial_config(self) -> int:
        self._seek(TRIAL_CONFIG_OFFSET)
        return self._read_packed('<H')

    def _calculate_block_size(self):
        sync_stamp = 9 * self.has_sync
        sample_size = sum([d.size for d in self._channel_dtypes])

        num_samples = int((BLOCK_LEN - sync_stamp) / sample_size)
        block_len = num_samples * sample_size + sync_stamp

        return num_samples, block_len

    def _read_sync_offset(self) -> Union[None, int]:
        # For this read operation we assume that every synchronization offset is immediately followed by a
        # timestamp as it is described in the manuals. We need to pair every sync offset with a timestamp for
        # interpolation at a later point in time.
        offset_sign_bool = self._read_packed('B')
        offset_sign = 1 - 2 * offset_sign_bool
        offset_mag = self._read_packed('<Q')

        if offset_mag == 2 ** 64 - 1:
            return None

        offset = offset_sign * offset_mag

        return offset

    def _read_sample(self) -> List:
        ch_values = []

        for ch, dtype in zip(self._channels, self._channel_dtypes):
            val_bin = self._read(dtype.size)
            ch_values.append(dtype.decode(val_bin))

        return ch_values

    def _read_data_block(self) -> Tuple[List[List], int]:
        sync_tuple = None
        samples = []

        try:
            if self.has_sync:
                sync_tuple = self._read_sync_offset()

            for i in range(self._samples_per_block):
                sample = self._read_sample()
                samples += [sample]
        except IOError:
            pass

        return samples, sync_tuple

    def _read_contents(self) -> Tuple[List, List[Tuple[int, int]]]:
        sync_offsets = []
        samples = []
        sample_ctr = 0

        self._seek(DATA_LOG_OFFSET)
        while True:
            block_samples, sync_offset = self._read_data_block()

            if sync_offset is not None:
                sync_offsets += [(sample_ctr, sync_offset)]

            samples += block_samples
            sample_ctr += len(block_samples)

            if len(block_samples) < self.samples_per_block:
                # We have reached EOF
                break

        return samples, sync_offsets

    def _read_exg_regs(self) -> Tuple[bytes, bytes]:
        self._seek(EXG_REG_OFFSET)

        reg1 = self._read(EXG_REG_LEN)
        reg2 = self._read(EXG_REG_LEN)
        return reg1, reg2

    def _read_triaxcal_params(self, offset: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        fmt = ">" + 6 * 'h' + 9 * 'b'

        self._seek(offset)
        calib_param_bytes = self._read(struct.calcsize(fmt))
        params_raw = struct.unpack(fmt, calib_param_bytes)

        offset = np.array(params_raw[:3])
        gain = np.diag(params_raw[3:6])
        alignment = np.array(params_raw[6:]).reshape((3, 3))

        return offset, gain, alignment

    def read_data(self):
        samples, sync_offsets = self._read_contents()

        samples_per_ch = list(zip(*samples))
        arr_per_ch = [np.array(s) for s in samples_per_ch]
        samples_dict = dict(zip(self._channels, arr_per_ch))

        if self.has_sync and len(sync_offsets) > 0:
            off_index, offset = list(zip(*sync_offsets))
            off_index_arr = np.array(off_index)
            offset_arr = np.array(offset)
            sync_data = (off_index_arr, offset_arr)
        else:
            sync_data = ((), ())

        return samples_dict, sync_data

    def get_exg_reg(self, chip_id: int) -> ExGRegister:
        reg_content = self._exg_regs[chip_id]
        return ExGRegister(reg_content)

    def get_triaxcal_params(self, sensor: ESensorGroup) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        offset = TRIAXCAL_FILE_OFFSET[sensor]
        sc_offset = TRIAXCAL_OFFSET_SCALING[sensor]
        sc_gain = TRIAXCAL_GAIN_SCALING[sensor]
        sc_alignment = TRIAXCAL_ALIGNMENT_SCALING[sensor]

        offset, gain, alignment = self._read_triaxcal_params(offset)
        return offset * sc_offset, gain * sc_gain, alignment * sc_alignment

    @property
    def sample_rate(self) -> int:
        return self._sr

    @property
    def block_size(self) -> int:
        return self._block_size

    @property
    def samples_per_block(self) -> int:
        return self._samples_per_block

    @property
    def enabled_sensors(self) -> List[ESensorGroup]:
        return self._sensors

    @property
    def enabled_channels(self) -> List[EChannelType]:
        return self._channels

    @property
    def has_global_clock(self) -> bool:
        return self._rtc_diff != 0x0

    @property
    def global_clock_diff(self) -> int:
        return self._rtc_diff

    @property
    def start_timestamp(self) -> int:
        return self._start_ts

    @property
    def has_sync(self) -> bool:
        return bit_is_set(self._trial_config, TRIAL_CONFIG_SYNC)

    @property
    def is_sync_master(self) -> bool:
        return bit_is_set(self._trial_config, TRIAL_CONFIG_MASTER)

    @property
    def exg_reg1(self) -> ExGRegister:
        return self.get_exg_reg(0)

    @property
    def exg_reg2(self) -> ExGRegister:
        return self.get_exg_reg(1)



from typing import Tuple
from unittest import TestCase

from pyshimmer import EFirmwareType, ShimmerDock
from pyshimmer.test_util import MockSerial


class DockAPITest(TestCase):

    @staticmethod
    def create_sot(flush: bool = False) -> Tuple[ShimmerDock, MockSerial]:
        mock = MockSerial()
        # noinspection PyTypeChecker
        dock = ShimmerDock(mock, flush_before_req=flush)

        return dock, mock

    def test_context_manager(self):
        dock, mock = self.create_sot()

        self.assertFalse(mock.test_closed)

        with dock:
            pass

        self.assertTrue(mock.test_closed)

    def test_unknown_start_char(self):
        dock, mock = self.create_sot()

        mock.test_put_read_data(b'\x25')
        self.assertRaises(IOError, dock.get_firmware_version)

    def test_bad_arg_response(self):
        dock, mock = self.create_sot()

        mock.test_put_read_data(b'\x24\xfd')
        self.assertRaises(IOError, dock.get_firmware_version)

    def test_bad_cmd_response(self):
        dock, mock = self.create_sot()

        mock.test_put_read_data(b'\x24\xfc')
        self.assertRaises(IOError, dock.get_firmware_version)

    def test_bad_crc_response(self):
        dock, mock = self.create_sot()

        mock.test_put_read_data(b'\x24\xfe')
        self.assertRaises(IOError, dock.get_firmware_version)

    def test_unexpected_cmd_response(self):
        dock, mock = self.create_sot()

        mock.test_put_read_data(b'\x24\x03')
        self.assertRaises(IOError, dock.get_firmware_version)

    def test_unexpected_component(self):
        dock, mock = self.create_sot()

        mock.test_put_read_data(b'\x24\x02\x02\x02\x00\x98z')
        self.assertRaises(IOError, dock.get_firmware_version)

    def test_unexpected_property(self):
        dock, mock = self.create_sot()

        mock.test_put_read_data(b'\x24\x02\x02\x01\x02\xaaE')
        self.assertRaises(IOError, dock.get_firmware_version)

    def test_get_mac_address(self):
        dock, mock = self.create_sot()

        mock.test_put_read_data(b'\x24\x02\x08\x01\x02\x01\x02\x03\x04\x05\x06N\x87')
        r = dock.get_mac_address()

        self.assertEqual(r, (0x01, 0x02, 0x03, 0x04, 0x05, 0x06))
        self.assertEqual(mock.test_get_write_data(), b'\x24\x03\x02\x01\x02\xfb\xef')

    def test_get_firmware_version(self):
        dock, mock = self.create_sot()

        mock.test_put_read_data(b'\x24\x02\x09\x01\x03\x03\x03\x00\x00\x00\x0b\x00\x14\x33')
        hw_ver, fw_type, major, minor, patch = dock.get_firmware_version()

        self.assertEqual(mock.test_get_write_data(), b'\x24\x03\x02\x01\x03\xca\xdc')

        self.assertEqual(hw_ver, 3)
        self.assertEqual(fw_type, EFirmwareType.LogAndStream)
        self.assertEqual(major, 0)
        self.assertEqual(minor, 11)
        self.assertEqual(patch, 0)

    def test_set_rtc(self):
        dock, mock = self.create_sot()

        mock.test_put_read_data(b'\x24\xff\xd9\xb2')
        dock.set_rtc(1.0)

        wd = mock.test_get_write_data()
        self.assertEqual(wd, b'\x24\x01\x0a\x01\x04\x00\x80\x00\x00\x00\x00\x00\x00\x1c\xd2')

    def test_get_rtc(self):
        dock, mock = self.create_sot()

        mock.test_put_read_data(b'\x24\x02\x0a\x01\x05\x9d\x3d\x0d\x00\x00\x00\x00\x00\xb0\xc7')
        r = dock.get_rtc()
        self.assertAlmostEqual(r, 26.481353759765625)

    def test_get_config_rtc(self):
        dock, mock = self.create_sot()

        mock.test_put_read_data(b'\x24\x02\x0a\x01\x04\x00\x00\x15\x00\x00\x00\x00\x00\xe4\xae')
        r = dock.get_config_rtc()
        self.assertEqual(r, 42.0)

        wd = mock.test_get_write_data()
        self.assertEqual(wd, b'\x24\x03\x02\x01\x04\x5d\x45')

    def test_get_exg_register(self):
        dock, mock = self.create_sot()

        # Due to the firmware bug, we first need to emulate the call to set the DAUGHTER_CARD CARD_ID
        mock.test_put_read_data(b'\x24\x02\x02\x03\x02\xca\x2b')
        exp_send_data1 = b'\x24\x03\x05\x03\x02\x00\x00\x00\x3a\xd2'

        # Then the actual call to retrieve the infomem data
        mock.test_put_read_data(b'\x24\x02\x0c\x01\x06\x00\x80\x10\x00\x00\x00\x00\x00\x02\x01\xff\x40')
        exp_send_data2 = b'\x24\x03\x05\x01\x06\x0a\x0a\x00\x42\x74'

        r = dock.get_exg_register(0)

        wd = mock.test_get_write_data()
        self.assertEqual(wd, exp_send_data1 + exp_send_data2)

        self.assertEqual(r.binary, b'\x00\x80\x10\x00\x00\x00\x00\x00\x02\x01')

    def test_get_exg_register_fail(self):
        dock, mock = self.create_sot()

        self.assertRaises(ValueError, dock.get_exg_register, -1)



# pyshimmer - API for Shimmer sensor devices
# Copyright (C) 2020  Lukas Magel

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import binascii
import struct

from serial import Serial

from pyshimmer.serial_base import SerialBase
from pyshimmer.uart.dock_const import CRC_INIT


def generate_crc(msg: bytes, crc_init: int) -> bytes:
    if len(msg) % 2 != 0:
        msg += b'\x00'

    crc = binascii.crc_hqx(msg, crc_init)
    crc_bin = struct.pack('<H', crc)
    return crc_bin


class DockSerial(SerialBase):

    def __init__(self, serial: Serial, crc_init: int = CRC_INIT):
        super().__init__(serial)
        self._crc_init = crc_init

        self._record_read = False
        self._read_crc_buf = bytearray()

        self._record_write = False
        self._write_crc_buf = bytearray()

    def _create_crc(self, msg: bytes) -> bytes:
        return generate_crc(msg, self._crc_init)

    def read(self, read_len: int) -> bytes:
        data = super().read(read_len)

        if self._record_read:
            self._read_crc_buf += data

        return data

    def write(self, data: bytes) -> int:
        if self._record_write:
            self._write_crc_buf += data

        return super().write(data)

    def start_read_crc_verify(self) -> None:
        self._record_read = True
        self._read_crc_buf = bytearray()

    def end_read_crc_verify(self) -> None:
        self._record_read = False

        exp_crc = self._create_crc(self._read_crc_buf)
        act_crc = super().read(len(exp_crc))
        if not exp_crc == act_crc:
            raise IOError('CRC check failed: Received data is invalid')

    def start_write_crc(self) -> None:
        self._record_write = True
        self._write_crc_buf = bytearray()

    def end_write_crc(self) -> None:
        self._record_write = False

        crc = self._create_crc(self._write_crc_buf)
        super().write(crc)



# pyshimmer - API for Shimmer sensor devices
# Copyright (C) 2020  Lukas Magel

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Tuple
from unittest import TestCase

from pyshimmer.test_util import MockSerial
from pyshimmer.uart.dock_serial import DockSerial, generate_crc


class CRCCalculationTest(TestCase):

    def test_generate_crc_uneven(self):
        crc_init = 0xB0CA
        msg = b'\x24\x03\x02\x01\x03'
        exp_crc = b'\xca\xdc'

        act_crc = generate_crc(msg, crc_init)
        self.assertEqual(act_crc, exp_crc)

    def test_generate_crc_even(self):
        crc_init = 0xB0CA
        msg = b'\x24\x03\x02\x01'
        exp_crc = b'\x4b\xc2'

        act_crc = generate_crc(msg, crc_init)
        print(act_crc)
        self.assertEqual(act_crc, exp_crc)


class DockSerialTest(TestCase):

    @staticmethod
    def create_sot(crc_init: int = 10) -> Tuple[DockSerial, MockSerial]:
        mock = MockSerial()
        # noinspection PyTypeChecker
        serial = DockSerial(mock, crc_init=crc_init)
        return serial, mock

    def test_read(self):
        crc_init = 42
        serial, mock = self.create_sot(crc_init)

        data_no_verify = b'abcd'
        data = b'\x01\x02\x03\x04'
        crc = generate_crc(data, crc_init)
        mock.test_put_read_data(data_no_verify + data + crc)

        r = serial.read(4)
        self.assertEqual(data_no_verify, r)

        serial.start_read_crc_verify()
        r = serial.read(4)
        serial.end_read_crc_verify()

        self.assertEqual(data, r)
        self.assertEqual(mock.test_get_remaining_read_data(), b'')

        mock.reset_input_buffer()
        mock.test_put_read_data(data + b'\x00\x01')

        serial.start_read_crc_verify()
        r = serial.read(4)
        self.assertEqual(r, data)
        self.assertRaises(IOError, serial.end_read_crc_verify)

    def test_write(self):
        crc_init = 42
        serial, mock = self.create_sot(crc_init)

        data_no_verify = b'1234'
        data = b'another test'
        crc = generate_crc(data, crc_init)

        serial.write(data_no_verify)

        serial.start_write_crc()
        serial.write(data)
        serial.end_write_crc()

        r = mock.test_get_write_data()
        self.assertEqual(len(r), 18)
        self.assertEqual(r[:4], b'1234')
        self.assertEqual(r[4:16], b'another test')
        self.assertEqual(r[-2:], crc)



# pyshimmer - API for Shimmer sensor devices
# Copyright (C) 2020  Lukas Magel

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import struct
from io import SEEK_SET, SEEK_CUR
from queue import Queue
from typing import BinaryIO, Tuple, Union, List

import numpy as np
import math


def bit_is_set(bitfield: int, mask: int) -> bool:
    """Check if the bit set in the mask is also set in the bitfield

    :param bitfield: The bitfield stored as an integer of arbitrary length
    :param mask: The mask where only a single bit is set
    :return: True if the bit in the mask is set in the bitfield, else False
    """
    return bitfield & mask == mask


def raise_to_next_pow(x: int) -> int:
    """Raise the argument to the next power of 2

    Example:
        - 1 --> 1
        - 2 --> 2
        - 3 --> 4
        - 5 --> 8

    :param x: The value to raise to the next power
    :return: The raised value
    """
    if x <= 0:
        return 1

    return 1 << (x - 1).bit_length()


def flatten_list(lst: Union[List, Tuple]) -> List:
    """Flatten the supplied list by one level

    Assumes that the supplied argument consists of lists itself. All elements are taken from the sublists and added
    to a fresh copy.

    :param lst: A list of lists
    :return: A list with the contents of the sublists
    """
    lst_flat = [val for sublist in lst for val in sublist]
    return lst_flat


def fmt_hex(val: bytes) -> str:
    """Format the supplied array of bytes as str

    :param val: The binary array to format
    :return: The resulting string
    """
    return ' '.join('{:02x}'.format(i) for i in val)


def unpack(args: Union[List, Tuple]) -> Union[List, Tuple, any]:
    """Extract the first object if the list has length 1

    If the supplied list or tuple only features a single element, the element is retrieved and returned. If the list or
    tuple is longer, the entire list or tuple is returned.

    :param args: The list or tuple to unpack
    :return: The list or tuple itself or the single element if the argument has a length of 1
    """
    if len(args) == 1:
        return args[0]
    return args


def unwrap(x: np.ndarray, shift: int) -> np.ndarray:
    """Detect overflows in the data and unwrap them

    The function tries to detect overflows in the input array x, with shape (N, ). It is assumed that x is monotonically
    increasing everywhere but at the overflows. An overflow occurs if for two consecutive points x_i and x_i+1 in the
    series x_i > x_i+1. For every such point, the function will add the value of the shift parameter to all following
    samples, i.e. x_k' = x_k + shift, for every k > i.

    :param x: The array to unwrap
    :param shift: The value which to add to the series after each overflow point
    :return: An array of equal length that has been unwrapped
    """
    x_diff = np.diff(x)
    wrap_points = np.argwhere(x_diff < 0)

    for i in wrap_points.flatten():
        x += (np.arange(len(x)) > i) * shift

    return x


def resp_code_to_bytes(code: Union[int, Tuple[int, ...], bytes]) -> bytes:
    """Convert the supplied response code to bytes

    :param code: The code, can be an int, a tuple of ints, or bytes
    :return: The supplied code as byte array
    """
    if isinstance(code, int):
        code = (code,)
    if isinstance(code, tuple):
        code = bytes(code)

    return code


def calibrate_u12_adc_value(uncalibratedData, offset, vRefP, gain):
    """Convert the uncalibrated data to calibrated data

    :param uncalibratedData: Raw voltage measurement from device
    :param offset: Voltage offset in measured data
    :param vRefP: Voltage reference signal in Volt
    :param gain: gain factor
    :return: Calibrated voltage in Volt
    """
    return ((uncalibratedData - offset) * ((vRefP/gain)/4095))


def battery_voltage_to_percent(battery_voltage):
    """Convert battery voltage to percent

    :param battery_voltage: Battery voltage in Volt
    :return: approximated battery state in percent based on manual 
    """
    # reference values from: https://shimmersensing.com/wp-content/docs/support/documentation/Shimmer_User_Manual_rev3p.pdf (Page 53)
    reference_data_voltages = [3.2, 3.627, 3.645, 3.663, 3.681, 3.699, 3.717, 3.7314, 3.735, 3.7386, 3.7566, 3.771, 3.789, 3.8034, 3.8106, 3.8394, 3.861, \
        3.8826, 3.9078, 3.933, 3.969, 4.0086, 4.041, 4.0734, 4.113, 4.167]
    reference_data_percentages = [0, 5.9, 9.8, 13.8, 17.7, 21.6, 25.6, 29.5, 33.4, 37.4, 41.3, 45.2, 49.2, 53.1, 57, 61, 64.9, 68.9, 72.8, 76.7, 80.7, 84.6, \
        88.5, 92.5, 96.4, 100]

    battery_percent = np.interp(battery_voltage, reference_data_voltages, reference_data_percentages)

    battery_percent = min(battery_percent, 100)
    battery_percent = max(battery_percent, 0)

    return battery_percent


class PeekQueue(Queue):
    """A thread-safe queue implementation that allows peeking at the first element in the queue.

    Based on a suggestion on StackOverflow:
    https://stackoverflow.com/questions/1293966/best-way-to-obtain-indexed-access-to-a-python-queue-thread-safe
    """

    def peek(self):
        """Peek at the element that will be removed next.

        :return: The next entry in the queue to be removed or None if the queue is empty
        """
        # noinspection PyUnresolvedReferences
        with self.mutex:
            if self._qsize() > 0:
                return self.queue[0]

            return None


class FileIOBase:
    """Convenience wrapper around a BinaryIO file object

    Serves as an (abstract) base class for IO operations

    :arg fp: The file to wrap
    """

    def __init__(self, fp: BinaryIO):
        if not fp.seekable():
            raise ValueError('IO object must be seekable')

        self._fp = fp

    def _tell(self) -> int:
        return self._fp.tell()

    def _read(self, s: int) -> bytes:
        r = self._fp.read(s)
        if len(r) < s:
            raise IOError('Read beyond EOF')

        return r

    def _seek(self, off: int = 0) -> None:
        self._fp.seek(off, SEEK_SET)

    def _seek_relative(self, off: int = 0) -> None:
        self._fp.seek(off, SEEK_CUR)

    def _read_packed(self, fmt: str) -> any:
        s = struct.calcsize(fmt)
        val_bin = self._read(s)

        args = struct.unpack(fmt, val_bin)
        return unpack(args)



# pyshimmer - API for Shimmer sensor devices
# Copyright (C) 2023  Lukas Magel, Manuel Fernandez-Carmona

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import struct
from typing import List

from pyshimmer.util import fmt_hex


class AllCalibration:

    def __init__(self, reg_bin: bytes):
        self._num_bytes = 84
        self._sensor_bytes = 21
        self._num_sensors = 4

        if len(reg_bin) < self._num_bytes:
            raise ValueError(
                f'All calibration data must have length {self._num_bytes}')

        self._reg_bin = reg_bin

    def __str__(self) -> str:
        def print_sensor(sens_num: int) -> str:
            return f'Sensor {sens_num + 1:2d}\n' + \
                   f'\tOffset bias: {self.get_offset_bias(sens_num)}\n' + \
                   f'\tSensitivity: {self.get_sensitivity(sens_num)}\n' + \
                   f'\tAlignment Matrix: {self.get_ali_mat(sens_num)}\n'

        obj_str = f''
        for i in range(0, self._num_sensors):
            obj_str += print_sensor(i)

        reg_bin_str = fmt_hex(self._reg_bin)
        obj_str += f'Binary: {reg_bin_str}\n'

        return obj_str

    @property
    def binary(self):
        return self._reg_bin
    
    def __eq__(self, other: "AllCalibration") -> bool:
        return self._reg_bin == other._reg_bin

    def _check_sens_num(self, sens_num: int) -> None:
        if not 0 <= sens_num < (self._num_sensors):
            raise ValueError(f'Sensor num must be 0 to {self._num_sensors-1}')

    def get_offset_bias(self, sens_num: int) -> List[int]:
        self._check_sens_num(sens_num)
        start_offset = sens_num * self._sensor_bytes
        end_offset = start_offset + 6
        ans = list(struct.unpack(
            '>hhh', self._reg_bin[start_offset:end_offset]))
        return ans

    def get_sensitivity(self, sens_num: int) -> List[int]:
        self._check_sens_num(sens_num)
        start_offset = sens_num * self._sensor_bytes + 6
        end_offset = start_offset + 6
        ans = list(struct.unpack(
            '>hhh', self._reg_bin[start_offset:end_offset]))
        return ans

    def get_ali_mat(self, sens_num: int) -> List[int]:
        self._check_sens_num(sens_num)
        start_offset = sens_num * self._sensor_bytes + 12
        end_offset = start_offset + 9
        ans = list(struct.unpack(
            '>bbbbbbbbb', self._reg_bin[start_offset:end_offset]))
        return ans



# pyshimmer - API for Shimmer sensor devices
# Copyright (C) 2020  Lukas Magel

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import struct
from io import RawIOBase
from typing import Callable

from serial import Serial

from pyshimmer.util import unpack


class ReadAbort(Exception):
    """
    Raised by ShimmerSerial when a read operation was cancelled by the user.
    """
    pass


class BufferedReader:
    """
    Wraps around a RawIOBase object to enable buffered reads and peeks while only requesting as many bytes as needed
    for the current request from the underlying stream.

    :param io_obj: The binary stream from which to read data
    """

    def __init__(self, io_obj: RawIOBase):
        self._io_obj = io_obj
        self._buf = bytearray()

    def _do_read_or_throw(self, read_len: int) -> bytes:
        result = self._io_obj.read(read_len)
        if len(result) < read_len:
            raise ReadAbort('Read operation returned prematurely. Read was cancelled.')
        return result

    def _fill_buffer(self, n: int) -> None:
        avail = len(self._buf)
        needed = n - avail

        if needed > 0:
            self._buf += self._do_read_or_throw(needed)

    def _get_from_buf(self, n: int) -> bytes:
        self._fill_buffer(n)
        return bytes(self._buf[:n])

    def _take_from_buf(self, n: int) -> bytes:
        data = self._get_from_buf(n)
        self._buf = self._buf[n:]
        return data

    def read(self, n: int) -> bytes:
        """Read n bytes from the underlying stream

        :param n: The number of bytes to read
        :raises ReadAbort: If the read operation on the underlying object has been cancelled and the stream returns
            less bytes than requested.
        :return: The requested bytes
        """
        return self._take_from_buf(n)

    def peek(self, n: int) -> bytes:
        """Peek n bytes into the stream, but do not discard the data

        Takes n bytes from the stream and returns them but also keeps them in an internal buffer, such that they
        can be returned by a later call to :meth:`read`.

        :param n: The number of bytes to peek
        :raises ReadAbort: If the read operation on the underlying object has been cancelled and the stream returns
            less bytes than requested.
        :return: The requested bytes
        """
        return self._get_from_buf(n)

    def reset(self) -> None:
        """Clear the internal buffer of the reader

        """
        self._buf = bytearray()


class SerialBase:
    """Wrapper around :class:`serial.Serial` which provides multiple convenience methods.

    :param serial: The serial instance to wrap
    """

    def __init__(self, serial: Serial):
        if serial.timeout is not None:
            print('Warning: Serial Read timeout != None. This will interfere with the detection cancelled reads.')

        self._serial = serial
        self._reader = BufferedReader(serial)

    @staticmethod
    def _retrieve_packed(fn_read: Callable[[int], bytes], rformat: str) -> any:
        read_len = struct.calcsize(rformat)

        r = fn_read(read_len)
        args_unpacked = struct.unpack(rformat, r)
        return unpack(args_unpacked)

    def flush_input_buffer(self):
        """Flush the input buffer and remove any data that has been received but not yet read

        """
        self._serial.reset_input_buffer()
        self._reader.reset()

    def write(self, data: bytes) -> int:
        """Write the data to the underlying serial stream

        The call blocks until the data has been written.

        :param data: The data to write
        :return: The number of bytes written, is equal to the length of data
        """
        return self._serial.write(data)

    def write_byte(self, arg: int) -> int:
        """Write a single byte to the stream

        :param arg: The byte to write
        :return: The number of bytes written
        """
        return self.write_packed('B', arg)

    def write_packed(self, wformat: str, *args) -> int:
        """Pack a number of arguments using :mod:`struct` and write them to the stream

        :param wformat: The format to use, see :func:`struct.pack`
        :param args: The arguments for the format string
        :return: The number of bytes written to the stream
        """
        args_packed = struct.pack(wformat, *args)
        return self.write(args_packed)

    def read(self, read_len: int) -> bytes:
        """Read the requested number of bytes from the stream

         The call blocks until sufficient data is available.

        :param read_len: The number of bytes to read
        :return: The data as bytes instance
        """
        return self._reader.read(read_len)

    def read_packed(self, rformat: str) -> any:
        """Read the data requested by the format string and unpack it

        The function evaluates the :mod:`struct` format string, reads the required number of bytes, and returns the
        unpacked arguments.

        :param rformat: The string format
        :return: A variable number of arguments, depending on the format string
        """
        return self._retrieve_packed(self.read, rformat)

    def read_byte(self) -> int:
        """Read a single byte from the stream

        :return: The byte that was read
        """
        r = self.read_packed('B')
        return r

    def cancel_read(self) -> None:
        """Cancel ongoing read operation

        If another thread is blocked in a read operation, cancel the operation by calling this method.
        """
        self._serial.cancel_read()

    def peek(self, n: int) -> bytes:
        """Peek the requested number of bytes into the stream

        :param n: The number of bytes to return
        :return: The requested bytes
        """
        return self._reader.peek(n)

    def peek_packed(self, rformat: str) -> any:
        """Peek into the stream and return the requested packed arguments

        :param rformat: The format of the data to peek
        :return: The peeked data, can be a variable number of arguments, depending on the format string
        """
        return self._retrieve_packed(self.peek, rformat)

    def close(self) -> None:
        """Close the underlying serial stream

        """
        self._serial.close()



# pyshimmer - API for Shimmer sensor devices
# Copyright (C) 2020  Lukas Magel

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from io import BytesIO
from typing import Tuple
from unittest import TestCase

from pyshimmer.test_util import MockSerial
from pyshimmer.serial_base import SerialBase, BufferedReader, ReadAbort


class BufferedReaderTest(TestCase):

    def test_read(self):
        stream = BytesIO(b'thisisatest')

        # noinspection PyTypeChecker
        reader = BufferedReader(stream)

        r = reader.read(0)
        self.assertEqual(r, b'')

        r = reader.read(1)
        self.assertEqual(r, b't')

        r = reader.read(3)
        self.assertEqual(r, b'his')

        r = reader.read(0)
        self.assertEqual(r, b'')

        r = reader.read(7)
        self.assertEqual(r, b'isatest')

        self.assertRaises(ReadAbort, reader.read, 1)

    def test_peek(self):
        stream = BytesIO(b'thisisatest')

        # noinspection PyTypeChecker
        reader = BufferedReader(stream)

        r = reader.peek(0)
        self.assertEqual(r, b'')

        for i in range(2):
            r = reader.peek(4)
            self.assertEqual(r, b'this')

        r = reader.peek(11)
        self.assertEqual(r, b'thisisatest')

        self.assertRaises(ReadAbort, reader.peek, 20)

    def test_peek_read(self):
        stream = BytesIO(b'thisisatest')

        # noinspection PyTypeChecker
        reader = BufferedReader(stream)

        r = reader.peek(4)
        self.assertEqual(r, b'this')

        r = reader.read(2)
        self.assertEqual(r, b'th')

        r = reader.peek(6)
        self.assertEqual(r, b'isisat')

        r = reader.read(3)
        self.assertEqual(r, b'isi')

    def test_reset(self):
        stream = BytesIO(b'thisisatest')

        # noinspection PyTypeChecker
        reader = BufferedReader(stream)

        r = reader.read(2)
        self.assertEqual(r, b'th')

        r = reader.peek(4)
        self.assertEqual(r, b'isis')

        reader.reset()
        r = reader.read(3)
        self.assertEqual(r, b'ate')

        reader.reset()
        r = reader.peek(2)
        self.assertEqual(r, b'st')


class SerialBaseTest(TestCase):

    @staticmethod
    def create_sot() -> Tuple[MockSerial, SerialBase]:
        mock = MockSerial()

        # noinspection PyTypeChecker
        sot = SerialBase(mock)

        return mock, sot

    def test_flush_input_buf(self):
        mock, sot = self.create_sot()

        mock.test_put_read_data(b'test')
        r = sot.read(2)
        self.assertEqual(r, b'te')

        sot.flush_input_buffer()
        self.assertTrue(mock.test_input_flushed)
        self.assertRaises(ReadAbort, sot.read, 1)

    def test_write(self):
        mock, sot = self.create_sot()

        i = sot.write(b'this')
        self.assertEqual(i, 4)
        i = sot.write(b'')
        self.assertEqual(i, 0)
        i = sot.write(b'is')
        self.assertEqual(i, 2)

        r = mock.test_get_write_data()
        self.assertEqual(r, b'thisis')

        i = sot.write_byte(10)
        self.assertEqual(i, 1)
        i = sot.write_byte(16)
        self.assertEqual(i, 1)

        r = mock.test_get_write_data()
        self.assertEqual(r, b'\x0A\x10')

        i = sot.write_packed('<HBI', 0x0100, 0x20, 0x30000)
        self.assertEqual(i, 7)

        r = mock.test_get_write_data()
        self.assertEqual(r, b'\x00\x01\x20\x00\x00\x03\x00')

    def test_read(self):
        mock, sot = self.create_sot()

        mock.test_put_read_data(b'another test')

        r = sot.read(0)
        self.assertEqual(r, b'')

        r = sot.read(4)
        self.assertEqual(r, b'anot')

        mock.test_clear_read_buffer()
        mock.test_put_read_data(b'\x10')

        r = sot.read_byte()
        self.assertEqual(r, 16)

        mock.test_put_read_data(b'\x42')
        r = sot.read_packed('B')
        self.assertEqual(r, 0x42)

        mock.test_put_read_data(b'\x00\x10\x30\x00\x00\x31\x00')
        r = sot.read_packed('<HBI')
        self.assertEqual(r, (0x1000, 0x30, 0x310000))

    def test_peek(self):
        mock, sot = self.create_sot()

        mock.test_put_read_data(b'hello' + b'\x30\x40\x01\x10')

        r = sot.peek(5)
        self.assertEqual(r, b'hello')

        r = sot.read(5)
        self.assertEqual(r, b'hello')

        r = sot.peek_packed('<B')
        self.assertEqual(r, 0x30)

        r = sot.peek_packed('<BH')
        self.assertEqual(r, (0x30, 0x140))

        r = sot.read_packed('<H')
        self.assertEqual(r, 0x4030)

    def test_close(self):
        mock, sot = self.create_sot()

        sot.close()
        self.assertTrue(mock.test_closed)

    def test_cancel_read(self):
        mock, sot = self.create_sot()

        sot.cancel_read()
        self.assertTrue(mock.test_read_cancelled)



# pyshimmer - API for Shimmer sensor devices
# Copyright (C) 2020  Lukas Magel

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import os
import pty
from io import BytesIO, RawIOBase, SEEK_END, SEEK_SET
from typing import Optional, Union, Tuple, BinaryIO

from serial import Serial


class MockSerial(RawIOBase):

    def __init__(self, timeout=None):
        self._read_buf = BytesIO()
        self._write_buf = BytesIO()

        self.timeout = timeout

        self.test_closed = False
        self.test_input_flushed = False
        self.test_read_cancelled = False

    def readable(self) -> bool:
        return True

    def writable(self) -> bool:
        return True

    def close(self) -> None:
        super().close()

        self.test_closed = True

    def readinto(self, b: bytearray) -> Optional[int]:
        return self._read_buf.readinto(b)

    def write(self, b: Union[bytes, bytearray]) -> Optional[int]:
        return self._write_buf.write(b)

    def reset_input_buffer(self):
        self.test_input_flushed = True

        self._read_buf.seek(0, SEEK_END)

    def cancel_read(self):
        self.test_read_cancelled = True

    def test_put_read_data(self, data: bytes) -> None:
        cur_pos = self._read_buf.tell()

        self._read_buf.seek(0, SEEK_END)
        self._read_buf.write(data)
        self._read_buf.seek(cur_pos, SEEK_SET)

    def test_get_remaining_read_data(self) -> bytes:
        return self._read_buf.read()

    def test_clear_read_buffer(self) -> None:
        self._read_buf = BytesIO()

    def test_get_write_data(self) -> bytes:
        self._write_buf.seek(0, SEEK_SET)
        data = self._write_buf.read()

        self._write_buf = BytesIO()
        return data


class PTYSerialMockCreator:

    def __init__(self):
        self._master_fobj = None
        self._slave_fobj = None

        self._slave_serial = None

    @staticmethod
    def _create_fobj(fd: int) -> BinaryIO:
        # https://bugs.python.org/issue20074
        fobj = os.fdopen(fd, 'r+b', 0)
        assert fobj.fileno() == fd

        return fobj

    def create_mock(self) -> Tuple[Serial, BinaryIO]:
        master_fd, slave_fd = pty.openpty()

        self._master_fobj = self._create_fobj(master_fd)
        self._slave_fobj = self._create_fobj(slave_fd)

        # Serial Baud rate is ignored by the driver and can be set to any value
        slave_path = os.ttyname(slave_fd)
        self._slave_serial = Serial(slave_path, 115200)

        return self._slave_serial, self._master_fobj

    def close(self):
        self._slave_serial.close()

        self._master_fobj.close()
        self._slave_fobj.close()



# pyshimmer - API for Shimmer sensor devices
# Copyright (C) 2020  Lukas Magel

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from pathlib import Path

_res_folder_name = 'resources'
_single_sample_name = 'single_sample.bin'
_synced_pair_bin_name = 'sdlog_sync_slave.bin'
_synced_pair_csv_name = 'sdlog_sync_slave.csv.gz'
_pair_raw_name = 'pair_raw.bin'
_pair_csv_name = 'pair_consensys.csv'

_acc_gyro_sample_name = "triaxcal_sample.bin"
_acc_gyro_uncal_name = "triaxcal_uncalibrated.csv.gz"
_acc_gyro_cal_name = "triaxcal_calibrated.csv.gz"

_ecg_sample_bin = 'ecg.bin'
_ecg_sample_uncal = 'ecg_uncalibrated.csv.gz'
_ecg_sample_cal = 'ecg_calibrated.csv.gz'


def get_resources_dir():
    my_dir = Path(__file__).parent
    res_dir = my_dir / _res_folder_name
    return res_dir


def get_binary_sample_fpath():
    return get_resources_dir() / _single_sample_name


def get_bin_vs_consensys_pair_fpath():
    res_dir = get_resources_dir()
    return res_dir / _pair_raw_name, res_dir / _pair_csv_name


def get_synced_bin_vs_consensys_pair_fpath():
    res_dir = get_resources_dir()
    return res_dir / _synced_pair_bin_name, res_dir / _synced_pair_csv_name


def get_ecg_sample():
    res_dir = get_resources_dir()
    return res_dir / _ecg_sample_bin, res_dir / _ecg_sample_uncal, res_dir / _ecg_sample_cal


def get_triaxcal_sample():
    res_dir = get_resources_dir()
    return res_dir / _acc_gyro_sample_name, res_dir / _acc_gyro_uncal_name, res_dir / _acc_gyro_cal_name



# pyshimmer - API for Shimmer sensor devices
# Copyright (C) 2020  Lukas Magel

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from unittest import TestCase

import numpy as np

from pyshimmer.dev.channels import ESensorGroup, get_ch_dtypes
from pyshimmer import EChannelType, ExGRegister
from pyshimmer.reader.shimmer_reader import ShimmerBinaryReader
from .reader_test_util import get_binary_sample_fpath, get_synced_bin_vs_consensys_pair_fpath, get_ecg_sample, \
    get_triaxcal_sample


class ShimmerReaderTest(TestCase):

    def test_parsing_wo_sync(self):
        fpath = get_binary_sample_fpath()
        with open(fpath, 'rb') as f:
            reader = ShimmerBinaryReader(f)

            exp_dr = 65
            exp_sensors = [ESensorGroup.ACCEL_LN, ESensorGroup.BATTERY, ESensorGroup.CH_A13]
            exp_channels = [EChannelType.TIMESTAMP, EChannelType.ACCEL_LN_X, EChannelType.ACCEL_LN_Y,
                            EChannelType.ACCEL_LN_Z, EChannelType.VBATT, EChannelType.INTERNAL_ADC_13]

            sample_size = sum([dt.size for dt in get_ch_dtypes(exp_channels)])
            samples_per_block = int(512 / sample_size)
            block_size = samples_per_block * sample_size

            self.assertEqual(reader.enabled_sensors, exp_sensors)
            self.assertEqual(reader.enabled_channels, exp_channels)
            self.assertEqual(reader.sample_rate, exp_dr)
            self.assertEqual(reader.has_global_clock, True)
            self.assertEqual(reader.has_sync, False)
            self.assertEqual(reader.is_sync_master, False)
            self.assertEqual(reader.samples_per_block, samples_per_block)
            self.assertEqual(reader.start_timestamp, 31291951)
            self.assertEqual(reader.block_size, block_size)
            self.assertEqual(reader.exg_reg1.binary, b'\x00\x80\x10\x00\x00\x00\x00\x00\x02\x01')
            self.assertEqual(reader.exg_reg2.binary, b'\x00\x80\x10\x00\x00\x00\x00\x00\x02\x01')

            data, _ = reader.read_data()
            ts = data[EChannelType.TIMESTAMP]

            # Sanity check on the timestamps: they should all be spaced equally apart with a stride that is equal
            # to the sampling rate.
            ts_diff = np.diff(ts)
            correct_diff = np.sum(ts_diff == exp_dr)
            self.assertTrue(correct_diff / len(ts_diff) > 0.98)

    def test_parsing_w_sync(self):
        fpath, _ = get_synced_bin_vs_consensys_pair_fpath()
        with open(fpath, 'rb') as f:
            reader = ShimmerBinaryReader(f)

            exp_dr = 64
            exp_sensors = [ESensorGroup.CH_A13]
            exp_channels = [EChannelType.TIMESTAMP, EChannelType.INTERNAL_ADC_13]
            exp_offsets = np.array([372, 362, 364, 351])
            exp_sync_ts = np.array([3725366, 4071094, 4397558, 4724022])
            exp_exg_reg1 = ExGRegister(b'\x00\x80\x10\x00\x00\x00\x00\x00\x02\x01')
            exp_exg_reg2 = ExGRegister(b'\x00\x80\x10\x00\x00\x00\x00\x00\x02\x01')

            sample_size = sum([dt.size for dt in get_ch_dtypes(exp_channels)])
            samples_per_block = int((512 - 9) / sample_size)
            block_size = samples_per_block * sample_size + 9

            self.assertEqual(reader.has_global_clock, True)
            self.assertEqual(reader.has_sync, True)
            self.assertEqual(reader.is_sync_master, False)
            self.assertEqual(reader.enabled_sensors, exp_sensors)
            self.assertEqual(reader.enabled_channels, exp_channels)
            self.assertEqual(reader.sample_rate, exp_dr)
            self.assertEqual(reader.samples_per_block, samples_per_block)
            self.assertEqual(reader.start_timestamp, 3085110)
            self.assertEqual(reader.block_size, block_size)

            self.assertEqual(reader.get_exg_reg(0), exp_exg_reg1)
            self.assertEqual(reader.get_exg_reg(1), exp_exg_reg2)
            self.assertEqual(reader.exg_reg1, exp_exg_reg1)
            self.assertEqual(reader.exg_reg2, exp_exg_reg2)

            samples, (off_index, sync_off) = reader.read_data()
            ts = samples[EChannelType.TIMESTAMP]

            np.testing.assert_equal(ts[off_index], exp_sync_ts)
            np.testing.assert_equal(sync_off, exp_offsets)

            # Sanity check on the timestamps: they should all be spaced equally apart with a stride that is equal
            # to the sampling rate.
            ts = samples[EChannelType.TIMESTAMP]
            ts_diff = np.diff(ts)
            correct_diff = np.sum(ts_diff == exp_dr)
            self.assertTrue(correct_diff / len(ts_diff) > 0.98)

    def test_ecg_registers(self):
        fpath, _, _ = get_ecg_sample()
        with open(fpath, 'rb') as f:
            reader = ShimmerBinaryReader(f)
            self.assertEqual(reader.exg_reg1.binary, b'\x03\xA8\x10\x49\x40\x23\x00\x00\x02\x03')
            self.assertEqual(reader.exg_reg2.binary, b'\x03\xA0\x10\xC1\xC1\x00\x00\x00\x02\x01')

    # noinspection PyMethodMayBeStatic
    def test_accel_ln_calib_data(self):
        exp_params = {
            ESensorGroup.ACCEL_LN: (
                np.array([2045, 2071, 2033]),
                np.diag([83, 83, 83]),
                np.array([[0.0, 1.0, 0.0],
                          [1.0, 0.0, 0.02],
                          [0.02, -0.01, -1.0]]),
            ),
            ESensorGroup.GYRO: (
                np.array([-123, -29, -35]),
                np.diag([56.68, 57.91, 59.21]),
                np.array([[0.0, 1.0, -0.02],
                          [1.0, 0.0, 0.03],
                          [-0.25, 0.01, -0.97]])
            )
        }

        fpath, _, _ = get_triaxcal_sample()
        with open(fpath, 'rb') as f:
            reader = ShimmerBinaryReader(f)

            for sensor, (exp_offset, exp_gain, exp_alignment) in exp_params.items():
                offset, gain, alignment = reader.get_triaxcal_params(sensor)
                np.testing.assert_almost_equal(offset, exp_offset, decimal=10)
                np.testing.assert_almost_equal(gain, exp_gain, decimal=10)
                np.testing.assert_almost_equal(alignment, exp_alignment, decimal=10)



# pyshimmer - API for Shimmer sensor devices
# Copyright (C) 2023  Lukas Magel

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from enum import Enum, auto


def ensure_firmware_version(func):
    def wrapper(self, other):
        if not isinstance(other, FirmwareVersion):
            return False

        return func(self, other)

    return wrapper


class EFirmwareType(Enum):
    BtStream = auto()
    SDLog = auto()
    LogAndStream = auto()


class FirmwareVersion:

    def __init__(self, major: int, minor: int, rel: int):
        """Represents the version of the Shimmer firmware

        :param major: Major version
        :param minor: Minor version
        :param rel: Patch level
        """
        self.major = major
        self.minor = minor
        self.rel = rel
        self._key = (major, minor, rel)

    @ensure_firmware_version
    def __eq__(self, other: "FirmwareVersion") -> bool:
        return self._key == other._key

    @ensure_firmware_version
    def __gt__(self, other: "FirmwareVersion") -> bool:
        return self._key > other._key

    @ensure_firmware_version
    def __ge__(self, other: "FirmwareVersion") -> bool:
        return self._key >= other._key

    @ensure_firmware_version
    def __lt__(self, other: "FirmwareVersion") -> bool:
        return self._key < other._key

    @ensure_firmware_version
    def __le__(self, other: "FirmwareVersion") -> bool:
        return self._key <= other._key


class FirmwareCapabilities:

    def __init__(self, fw_type: EFirmwareType, version: FirmwareVersion):
        self._fw_type = fw_type
        self._version = version

    @property
    def fw_type(self) -> EFirmwareType:
        return self._fw_type

    @property
    def version(self) -> FirmwareVersion:
        return self._version

    @property
    def supports_ack_disable(self) -> bool:
        return self._fw_type == EFirmwareType.LogAndStream and \
               self._version >= FirmwareVersion(major=0, minor=15, rel=4)


FirmwareTypeValueAssignment = {
    0x01: EFirmwareType.BtStream,
    0x02: EFirmwareType.SDLog,
    0x03: EFirmwareType.LogAndStream,
}


def get_firmware_type(f_type: int) -> EFirmwareType:
    if f_type not in FirmwareTypeValueAssignment:
        raise ValueError(f'Unknown firmware type: 0x{f_type:x}')

    return FirmwareTypeValueAssignment[f_type]



# pyshimmer - API for Shimmer sensor devices
# Copyright (C) 2023  Lukas Magel

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import random
from unittest import TestCase

from pyshimmer.dev.channels import EChannelType
from pyshimmer.dev.exg import is_exg_ch, get_exg_ch, ExGMux, ExGRLDLead, ERLDRef, ExGRegister


def randbytes(k: int) -> bytes:
    population = list(range(256))
    seq = random.choices(population, k=k)
    return bytes(seq)


class DeviceExGTest(TestCase):

    def test_get_exg_ch(self):
        self.assertEqual(get_exg_ch(EChannelType.EXG_ADS1292R_1_CH1_24BIT), (0, 0))
        self.assertEqual(get_exg_ch(EChannelType.EXG_ADS1292R_1_CH2_24BIT), (0, 1))
        self.assertEqual(get_exg_ch(EChannelType.EXG_ADS1292R_2_CH1_24BIT), (1, 0))
        self.assertEqual(get_exg_ch(EChannelType.EXG_ADS1292R_2_CH2_24BIT), (1, 1))

    def test_is_exg_ch(self):
        from itertools import product
        valid_ch = [EChannelType[f'EXG_ADS1292R_{i}_CH{j}_{k}BIT'] for i, j, k in product([1, 2], [1, 2], [16, 24])]

        for ch in EChannelType:
            self.assertEqual(is_exg_ch(ch), ch in valid_ch)


class ExGRegisterTest(TestCase):

    def setUp(self) -> None:
        random.seed(0x42)

    def test_exg_register_fail(self):
        self.assertRaises(ValueError, ExGRegister, bytes())

    def test_exg_register(self):
        reg1 = bytes([3, 160, 16, 64, 71, 0, 0, 0, 2, 1])
        reg2 = bytes([0, 171, 16, 21, 21, 0, 0, 0, 2, 1])

        exg_reg1 = ExGRegister(reg1)
        exg_reg2 = ExGRegister(reg2)

        self.assertEqual(exg_reg1.ch1_gain, 4)
        self.assertEqual(exg_reg1.ch2_gain, 4)
        self.assertEqual(exg_reg1.ch1_mux, ExGMux.NORMAL)
        self.assertEqual(exg_reg1.get_ch_mux_bin(0), 0b0000)
        self.assertEqual(exg_reg1.ch2_mux, ExGMux.RLD_DRM)
        self.assertEqual(exg_reg1.get_ch_mux_bin(1), 0b0111)
        self.assertEqual(exg_reg1.ch1_powerdown, False)
        self.assertEqual(exg_reg1.ch2_powerdown, False)
        self.assertEqual(exg_reg1.data_rate, 1000)
        self.assertEqual(exg_reg1.binary, reg1)

        self.assertEqual(exg_reg2.ch1_gain, 1)
        self.assertEqual(exg_reg2.ch2_gain, 1)
        self.assertEqual(exg_reg2.ch1_mux, ExGMux.TEST_SIGNAL)
        self.assertEqual(exg_reg2.ch2_mux, ExGMux.TEST_SIGNAL)
        self.assertEqual(exg_reg2.ch1_powerdown, False)
        self.assertEqual(exg_reg2.ch2_powerdown, False)
        self.assertEqual(exg_reg2.data_rate, 125)
        self.assertEqual(exg_reg2.binary, reg2)

        self.assertRaises(ValueError, exg_reg1.get_ch_mux, 2)
        self.assertRaises(ValueError, exg_reg1.get_ch_mux, -1)

    def test_exg_register_powerdown(self):
        pd = 0x1 << 7
        reg_bin = bytes([3, 160, 16, pd, pd, 0, 0, 0, 2, 1])
        reg = ExGRegister(reg_bin)

        self.assertEqual(reg.ch1_powerdown, True)
        self.assertEqual(reg.ch2_powerdown, True)

    def test_exg_register_rld_powerdown(self):
        pd = 0x01 << 5
        reg_bin = bytes([0, 0, 0, 0, 0, pd, 0, 0, 0, 0])
        reg = ExGRegister(reg_bin)

        self.assertEqual(reg.rld_powerdown, False)

    def test_exg_register_rld_channels(self):
        reg_bin = bytes([0x03, 0xA8, 0x10, 0x40, 0x40, 0x2D, 0x00, 0x00, 0x02, 0x03])
        reg = ExGRegister(reg_bin)
        self.assertEqual(reg.rld_channels, [ExGRLDLead.RLD1P, ExGRLDLead.RLD2P, ExGRLDLead.RLD2N])

        reg_bin = bytes([0x03, 0xA8, 0x10, 0x40, 0x40, 0x00, 0x00, 0x00, 0x02, 0x03])
        reg = ExGRegister(reg_bin)
        self.assertEqual(reg.rld_channels, [])

    def test_exg_register_rld_ref(self):
        reg_bin = bytes([0x03, 0xA8, 0x10, 0x40, 0x40, 0x2D, 0x00, 0x00, 0x02, 0x03])
        reg = ExGRegister(reg_bin)
        self.assertEqual(reg.rld_ref, ERLDRef.INTERNAL)

        reg_bin = bytes([0x03, 0xA8, 0x10, 0x40, 0x40, 0x2D, 0x00, 0x00, 0x02, 0x01])
        reg = ExGRegister(reg_bin)
        self.assertEqual(reg.rld_ref, ERLDRef.EXTERNAL)

    def test_exg_register_print(self):
        reg_bin = bytes([0x03, 0xA8, 0x10, 0x40, 0x40, 0x2D, 0x00, 0x00, 0x02, 0x03])
        reg = ExGRegister(reg_bin)

        str_repr = str(reg)
        self.assertTrue('Data Rate: 1000' in str_repr)
        self.assertTrue('RLD Powerdown: False' in str_repr)

    def test_equality_operator(self):
        def do_assert(a: bytes, b: bytes, result: bool) -> None:
            self.assertEqual(ExGRegister(a) == ExGRegister(b), result)

        x = randbytes(10)
        y = randbytes(10)

        do_assert(x, y, False)
        do_assert(x, x, True)
        do_assert(y, y, True)

        for i in range(len(x)):
            y = bytearray(x)
            y[i] = random.randrange(0, 256)
            do_assert(x, y, False)



import time

from serial import Serial

from pyshimmer import ShimmerBluetooth, DEFAULT_BAUDRATE, DataPacket


def stream_cb(pkt: DataPacket) -> None:   
    print(f'Received new data packet: ') 
    for chan in pkt.channels:
        print(f'channel: ' + str(chan)) 
        print(f'value: ' + str(pkt[chan]))     
    print('') 

def main(args=None):    
    serial = Serial('/dev/rfcomm42', DEFAULT_BAUDRATE)
    shim_dev = ShimmerBluetooth(serial)

    shim_dev.initialize()

    dev_name = shim_dev.get_device_name()
    print(f'My name is: {dev_name}')

    info = shim_dev.get_firmware_version() 
    print("- firmware: [" + str(info[0]) + "]")
    print("- version: [" + str(info[1].major) + "." + str(info[1].minor) + "." + str(info[1].rel) + "]")
    
    shim_dev.add_stream_callback(stream_cb)

    shim_dev.start_streaming()
    time.sleep(5.0)
    shim_dev.stop_streaming()

    shim_dev.shutdown()


if __name__ == '__main__':
    main()


# pyshimmer - API for Shimmer sensor devices
# Copyright (C) 2023  Lukas Magel

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from unittest import TestCase

from pyshimmer.dev.fw_version import FirmwareVersion, get_firmware_type, EFirmwareType, FirmwareCapabilities


class DeviceFirmwareVersionTest(TestCase):

    def test_get_firmware_type(self):
        r = get_firmware_type(0x01)
        self.assertEqual(r, EFirmwareType.BtStream)
        r = get_firmware_type(0x02)
        self.assertEqual(r, EFirmwareType.SDLog)
        r = get_firmware_type(0x03)
        self.assertEqual(r, EFirmwareType.LogAndStream)

        self.assertRaises(ValueError, get_firmware_type, 0xFF)


class FirmwareCapabilitiesTest(TestCase):

    def test_capabilities(self):
        cap = FirmwareCapabilities(EFirmwareType.LogAndStream, version=FirmwareVersion(1, 2, 3))
        self.assertTrue(cap.supports_ack_disable)
        self.assertEqual(cap.version, FirmwareVersion(1, 2, 3))
        self.assertEqual(cap.fw_type, EFirmwareType.LogAndStream)


class FirmwareVersionTest(TestCase):

    def test_version_equality(self):
        a = FirmwareVersion(1, 2, 3)
        b = FirmwareVersion(1, 2, 3)
        c = FirmwareVersion(3, 2, 1)

        self.assertEqual(a, a)
        self.assertEqual(a, b)

        self.assertNotEqual(a, None)
        self.assertNotEqual(a, False)
        self.assertNotEqual(a, 10)
        self.assertNotEqual(a, c)

    def test_attributes(self):
        ver = FirmwareVersion(1, 2, 3)
        self.assertEqual(ver.major, 1)
        self.assertEqual(ver.minor, 2)
        self.assertEqual(ver.rel, 3)

    def test_greater_less(self):
        a = FirmwareVersion(3, 2, 1)

        b = FirmwareVersion(3, 2, 1)
        self.assertFalse(b > a)
        self.assertTrue(b >= a)
        self.assertFalse(b < a)
        self.assertTrue(b <= a)

        b = FirmwareVersion(2, 2, 1)
        self.assertFalse(b > a)
        self.assertFalse(b >= a)
        self.assertTrue(b < a)
        self.assertTrue(b <= a)

        b = FirmwareVersion(3, 1, 1)
        self.assertFalse(b > a)
        self.assertFalse(b >= a)
        self.assertTrue(b < a)
        self.assertTrue(b <= a)

        b = FirmwareVersion(3, 2, 0)
        self.assertFalse(b > a)
        self.assertFalse(b >= a)
        self.assertTrue(b < a)
        self.assertTrue(b <= a)



# pyshimmer - API for Shimmer sensor devices
# Copyright (C) 2020  Lukas Magel

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from pyshimmer.dev.channels import EChannelType

ACK_COMMAND_PROCESSED = 0xFF
INSTREAM_CMD_RESPONSE = 0x8A
DATA_PACKET = 0x00

INQUIRY_COMMAND = 0x01
INQUIRY_RESPONSE = 0x02

GET_SAMPLING_RATE_COMMAND = 0x03
SAMPLING_RATE_RESPONSE = 0x04

SET_SAMPLING_RATE_COMMAND = 0x05

GET_BATTERY_COMMAND = 0x95
BATTERY_RESPONSE = 0x94
FULL_BATTERY_RESPONSE = bytes((INSTREAM_CMD_RESPONSE, BATTERY_RESPONSE))

START_STREAMING_COMMAND = 0x07
# No response for command

SET_SENSORS_COMMAND = 0x08
# No response for command

STOP_STREAMING_COMMAND = 0x20
# No response for command

GET_CONFIGTIME_COMMAND = 0x87
CONFIGTIME_RESPONSE = 0x86

SET_CONFIGTIME_COMMAND = 0x85
# No response for set command

GET_RWC_COMMAND = 0x91
RWC_RESPONSE = 0x90

SET_RWC_COMMAND = 0x8F
# No response for set command

GET_STATUS_COMMAND = 0x72
STATUS_RESPONSE = 0x71
FULL_STATUS_RESPONSE = bytes((INSTREAM_CMD_RESPONSE, STATUS_RESPONSE))

GET_FW_VERSION_COMMAND = 0x2E
FW_VERSION_RESPONSE = 0x2F

GET_EXG_REGS_COMMAND = 0x63
EXG_REGS_RESPONSE = 0x62

SET_EXG_REGS_COMMAND = 0x61
# No response for set command

GET_EXPID_COMMAND = 0x7E
EXPID_RESPONSE = 0x7D

SET_EXPID_COMMAND = 0x7C
# No response for set command

GET_SHIMMERNAME_COMMAND = 0x7B
SHIMMERNAME_RESPONSE = 0x7A

SET_SHIMMERNAME_COMMAND = 0x79
# No response for set command

DUMMY_COMMAND = 0x96

START_LOGGING_COMMAND = 0x92
STOP_LOGGING_COMMAND = 0x93

ENABLE_STATUS_ACK_COMMAND = 0xA3

GET_ALL_CALIBRATION_COMMAND = 0x2C
ALL_CALIBRATION_RESPONSE = 0x2D

"""
The Bluetooth LogAndStream API assigns a numerical index to each channel type. This dictionary maps each index to the
corresponding channel type.
"""
BtChannelsByIndex = {
    0x00: EChannelType.ACCEL_LN_X,
    0x01: EChannelType.ACCEL_LN_Y,
    0x02: EChannelType.ACCEL_LN_Z,
    0x03: EChannelType.VBATT,
    0x04: EChannelType.ACCEL_LSM303DLHC_X,
    0x05: EChannelType.ACCEL_LSM303DLHC_Y,
    0x06: EChannelType.ACCEL_LSM303DLHC_Z,
    0x07: EChannelType.MAG_LSM303DLHC_X,
    0x08: EChannelType.MAG_LSM303DLHC_Y,
    0x09: EChannelType.MAG_LSM303DLHC_Z,
    0x0A: EChannelType.GYRO_MPU9150_X,
    0x0B: EChannelType.GYRO_MPU9150_Y,
    0x0C: EChannelType.GYRO_MPU9150_Z,
    0x0D: EChannelType.EXTERNAL_ADC_7,
    0x0E: EChannelType.EXTERNAL_ADC_6,
    0x0F: EChannelType.EXTERNAL_ADC_15,
    0x10: EChannelType.INTERNAL_ADC_1,
    0x11: EChannelType.INTERNAL_ADC_12,
    0x12: EChannelType.INTERNAL_ADC_13,
    0x13: EChannelType.INTERNAL_ADC_14,
    0x14: EChannelType.ACCEL_MPU9150_X,
    0x15: EChannelType.ACCEL_MPU9150_Y,
    0x16: EChannelType.ACCEL_MPU9150_Z,
    0x17: EChannelType.MAG_MPU9150_X,
    0x18: EChannelType.MAG_MPU9150_Y,
    0x19: EChannelType.MAG_MPU9150_Z,
    0x1A: EChannelType.TEMP_BMPX80,
    0x1B: EChannelType.PRESSURE_BMPX80,
    0x1C: EChannelType.GSR_RAW,
    0x1D: EChannelType.EXG_ADS1292R_1_STATUS,
    0x1E: EChannelType.EXG_ADS1292R_1_CH1_24BIT,
    0x1F: EChannelType.EXG_ADS1292R_1_CH2_24BIT,
    0x20: EChannelType.EXG_ADS1292R_2_STATUS,
    0x21: EChannelType.EXG_ADS1292R_2_CH1_24BIT,
    0x22: EChannelType.EXG_ADS1292R_2_CH2_24BIT,
    0x23: EChannelType.EXG_ADS1292R_1_CH1_16BIT,
    0x24: EChannelType.EXG_ADS1292R_1_CH2_16BIT,
    0x25: EChannelType.EXG_ADS1292R_2_CH1_16BIT,
    0x26: EChannelType.EXG_ADS1292R_2_CH2_16BIT,
    0x27: EChannelType.STRAIN_HIGH,
    0x28: EChannelType.STRAIN_LOW,
}



# pyshimmer - API for Shimmer sensor devices
# Copyright (C) 2023  Lukas Magel

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Union, overload

import numpy as np

# Device clock rate in ticks per second
DEV_CLOCK_RATE: float = 32768.0

DEFAULT_BAUDRATE = 115200


def sr2dr(sr: float) -> int:
    """Calculate equivalent device-specific rate for a sample rate in Hz

    Device-specific sample rates are given in absolute clock ticks per unit of time. This function can be used to
    calculate such a rate for the Shimmer3.

    Args:
        sr(float): The sampling rate in Hz

    Returns:
        An integer which represents the equivalent device-specific sampling rate
    """
    dr_dec = DEV_CLOCK_RATE / sr
    return round(dr_dec)


def dr2sr(dr: int):
    """Calculate equivalent sampling rate for a given device-specific rate

    Device-specific sample rates are given in absolute clock ticks per unit of time. This function can be used to
    calculate a regular sampling rate in Hz from such a rate.

    Args:
        dr(int): The absolute device rate as int

    Returns:
        A floating-point number that represents the sampling rate in Hz
    """
    return DEV_CLOCK_RATE / dr


@overload
def sec2ticks(t_sec: float) -> int: ...


@overload
def sec2ticks(t_sec: np.ndarray) -> np.ndarray: ...


def sec2ticks(t_sec: Union[float, np.ndarray]) -> Union[int, np.ndarray]:
    """Calculate equivalent device clock ticks for a time in seconds

    Args:
        t_sec: A time in seconds
    Returns:
        An integer which represents the equivalent number of clock ticks
    """
    return round(t_sec * DEV_CLOCK_RATE)


@overload
def ticks2sec(t_ticks: int) -> float: ...


@overload
def ticks2sec(t_ticks: np.ndarray) -> np.ndarray: ...


def ticks2sec(t_ticks: Union[int, np.ndarray]) -> Union[float, np.ndarray]:
    """Calculate the time in seconds equivalent to a device clock ticks count

    Args:
        t_ticks: A clock tick counter for which to calculate the time in seconds
    Returns:
        A floating point time in seconds that is equivalent to the number of clock ticks
    """
    return t_ticks / DEV_CLOCK_RATE



# pyshimmer - API for Shimmer sensor devices
# Copyright (C) 2020  Lukas Magel

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from unittest import TestCase
from pyshimmer.util import PeekQueue

import numpy as np

from unittest.mock import Mock

from io import BytesIO
from pyshimmer.util import bit_is_set, raise_to_next_pow, flatten_list, fmt_hex, unpack, unwrap, calibrate_u12_adc_value, battery_voltage_to_percent, \
     FileIOBase


class UtilTest(TestCase):

    def test_bit_is_set(self):
        r = bit_is_set(0x10, 0x01)
        self.assertEqual(r, False)

        r = bit_is_set(0x10, 0x10)
        self.assertEqual(r, True)

        r = bit_is_set(0x05, 0x01)
        self.assertEqual(r, True)

        r = bit_is_set(0x05, 0x02)
        self.assertEqual(r, False)

        r = bit_is_set(0x05, 0x04)
        self.assertEqual(r, True)

    def test_raise_to_next_pow(self):
        r = raise_to_next_pow(0)
        self.assertEqual(r, 1)

        r = raise_to_next_pow(1)
        self.assertEqual(r, 1)

        r = raise_to_next_pow(2)
        self.assertEqual(r, 2)

        r = raise_to_next_pow(3)
        self.assertEqual(r, 4)

        r = raise_to_next_pow(4)
        self.assertEqual(r, 4)

        r = raise_to_next_pow(6)
        self.assertEqual(r, 8)

        r = raise_to_next_pow(14)
        self.assertEqual(r, 16)

    def test_flatten_list(self):
        r = flatten_list([[10], [20]])
        self.assertEqual(r, [10, 20])

        r = flatten_list(((10,), (20,)))
        self.assertEqual(r, [10, 20])

        r = flatten_list([[10]])
        self.assertEqual(r, [10])

    def test_fmt_hex(self):
        r = fmt_hex(b'\x01')
        self.assertEqual(r, '01')

        r = fmt_hex(b'\x01\x02')
        self.assertEqual(r, '01 02')

    def test_unpack(self):
        r = unpack([10])

        self.assertEqual(r, 10)

        r = unpack([10, 20])
        self.assertEqual(r, [10, 20])

        r = unpack([])
        self.assertEqual(r, [])

        r = unpack(())
        self.assertEqual(r, ())

        r = unpack((10,))
        self.assertEqual(r, 10)

        r = unpack((10, 20))
        self.assertEqual(r, (10, 20))

    # noinspection PyMethodMayBeStatic
    def test_unwrap(self):
        shift = 10
        x = np.array([0, 1, 5, 8, 0, 2, 5, 10, 3, 7, 9])
        e = np.array([0, 1, 5, 8, 10, 12, 15, 20, 23, 27, 29])

        r = unwrap(x, shift)
        np.testing.assert_equal(r, e)

        x = np.array([0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2])
        e = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14])

        r = unwrap(x, 4)
        np.testing.assert_equal(r, e)

        e = np.arange(0, 2000, 45)
        x = e % 250

        r = unwrap(x, 250)
        np.testing.assert_equal(r, e)

        e = np.arange(0, 4 * (2 ** 24), 65)
        x = e % (2 ** 24)

        r = unwrap(x, 2 ** 24)
        np.testing.assert_equal(r, e)

    def test_calibrate_u12_adc_value(self):
        uncalibratedData = 2863
        offset = 0
        vRefP = 3.0
        gain = 1.0
        
        actual = calibrate_u12_adc_value(uncalibratedData, offset, vRefP, gain)

        desired = 2.0974358974358975
        np.testing.assert_almost_equal(actual, desired)

    def test_battery_voltage_to_percent(self):
        voltage = 3.9078
        desired = 72.8

        actual = battery_voltage_to_percent(voltage)
        np.testing.assert_equal(actual, desired)

    def test_peek_queue(self):
        queue = PeekQueue()

        queue.put(1)
        queue.put(2)
        queue.put(3)

        self.assertEqual(queue.peek(), 1)
        queue.get()

        self.assertEqual(queue.peek(), 2)
        queue.get()

        self.assertEqual(queue.peek(), 3)
        queue.get()

        self.assertEqual(queue.peek(), None)

    def test_file_io_base(self):
        input_bin = bytes(range(255))
        io_obj = BytesIO(input_bin)

        sut = FileIOBase(io_obj)
        self.assertEqual(sut._tell(), 0)

        sut._seek(10)
        self.assertEqual(sut._tell(), 10)

        sut._seek_relative(-2)
        self.assertEqual(sut._tell(), 8)

        r = sut._read(2)
        self.assertEqual(r, b'\x08\x09')

        r = sut._read_packed('<H')
        self.assertEqual(r, 0x0B0A)

    def test_file_io_base_not_seekable(self):
        mock = Mock(spec=BytesIO)
        mock.seekable.return_value = False

        self.assertRaises(ValueError, FileIOBase, mock)



# pyshimmer - API for Shimmer sensor devices
# Copyright (C) 2023  Lukas Magel, Manuel Fernandez-Carmona

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import random
from unittest import TestCase
from pyshimmer.dev.calibration import AllCalibration

def randbytes(k: int) -> bytes:
    population = list(range(256))
    seq = random.choices(population, k=k)
    return bytes(seq)

class AllCalibrationTest(TestCase):

    def test_equality_operator(self):
        def do_assert(a: bytes, b: bytes, result: bool) -> None:
            self.assertEqual(AllCalibration(a) == AllCalibration(b), result)

        x = randbytes(84)
        y = randbytes(84)

        do_assert(x, y, False)
        do_assert(x, x, True)
        do_assert(y, y, True)

        for i in range(len(x)):
            y = bytearray(x)
            y[i] = random.randrange(0, 256)
            do_assert(x, y, False)

    def setUp(self) -> None:
        random.seed(0x42)

    def test_allcalibration_fail(self):
        self.assertRaises(ValueError, AllCalibration, bytes())

    def test_allcalibration(self):
        bin_reg1 = bytes([0x08, 0xcd, 0x08, 0xcd, 0x08, 0xcd, 0x00, 0x5c, 0x00, 0x5c, 
                          0x00, 0x5c, 0x00, 0x9c, 0x00, 0x9c, 0x00, 0x00, 0x00, 0x00, 
                          0x9c, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x19, 0x96, 0x19, 
                          0x96, 0x19, 0x96, 0x00, 0x9c, 0x00, 0x9c, 0x00, 0x00, 0x00, 
                          0x00, 0x9c, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x9b, 
                          0x02, 0x9b, 0x02, 0x9b, 0x00, 0x9c, 0x00, 0x64, 0x00, 0x00, 
                          0x00, 0x00, 0x9c, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x06, 
                          0x87, 0x06, 0x87, 0x06, 0x87, 0x00, 0x9c, 0x00, 0x64, 0x00, 
                          0x00, 0x00, 0x00, 0x9c])

        bin_reg2 = bytes([0x08, 0xcd, 0x08, 0xcd, 0x08, 0xcd, 0x00, 0x5c, 0x00, 0x5c, 
                          0x00, 0x5c, 0x00, 0x9c, 0x00, 0x9c, 0x00, 0x00, 0x00, 0x00, 
                          0x9c, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x19, 0x96, 0x19, 
                          0x96, 0x19, 0x96, 0x00, 0x9c, 0x00, 0x9c, 0x00, 0x00, 0x00, 
                          0x00, 0x9c, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
                          0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
                          0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x06, 
                          0x87, 0x06, 0x87, 0x06, 0x87, 0x00, 0x9c, 0x00, 0x64, 0x00, 
                          0x00, 0x00, 0x00, 0x9c])

        allcalib1 = AllCalibration(bin_reg1)
        allcalib2 = AllCalibration(bin_reg2)
        self.assertEqual(allcalib1.get_offset_bias(0), [2253, 2253, 2253] )
        self.assertEqual(allcalib1.get_sensitivity(0), [92, 92, 92] )
        self.assertEqual(allcalib1.get_ali_mat(0),     [0, -100, 0, -100, 0, 0, 0, 0, -100] )
        self.assertEqual(allcalib1.get_offset_bias(1), [0, 0, 0] )
        self.assertEqual(allcalib1.get_sensitivity(1), [6550, 6550, 6550] )
        self.assertEqual(allcalib1.get_ali_mat(1),     [0, -100, 0, -100, 0, 0, 0, 0, -100] )
        self.assertEqual(allcalib1.get_offset_bias(2), [0, 0, 0] )
        self.assertEqual(allcalib1.get_sensitivity(2), [667, 667, 667] )
        self.assertEqual(allcalib1.get_ali_mat(2),     [0, -100, 0, 100, 0, 0, 0, 0, -100] )
        self.assertEqual(allcalib1.get_offset_bias(3), [0, 0, 0] )
        self.assertEqual(allcalib1.get_sensitivity(3), [1671, 1671, 1671] )
        self.assertEqual(allcalib1.get_ali_mat(3),     [0, -100, 0, 100, 0, 0, 0, 0, -100] )
        
        self.assertEqual(allcalib2.get_offset_bias(0), [2253, 2253, 2253])
        self.assertEqual(allcalib2.get_sensitivity(0), [92, 92, 92])
        self.assertEqual(allcalib2.get_ali_mat(0),     [0, -100, 0, -100, 0, 0, 0, 0, -100])
        self.assertEqual(allcalib2.get_offset_bias(1), [0, 0, 0])
        self.assertEqual(allcalib2.get_sensitivity(1), [6550, 6550, 6550])
        self.assertEqual(allcalib2.get_ali_mat(1),     [0, -100, 0, -100, 0, 0, 0, 0, -100])
        self.assertEqual(allcalib2.get_offset_bias(2), [0, 0, 0])
        self.assertEqual(allcalib2.get_sensitivity(2), [0, 0, 0])
        self.assertEqual(allcalib2.get_ali_mat(2),     [0, 0, 0, 0, 0, 0, 0, 0, 0])
        self.assertEqual(allcalib2.get_offset_bias(3), [0, 0, 0])
        self.assertEqual(allcalib2.get_sensitivity(3), [1671, 1671, 1671])
        self.assertEqual(allcalib2.get_ali_mat(3),     [0, -100, 0, 100, 0, 0, 0, 0, -100])

    def test_exg_register_print(self):
        bin_reg = bytes([0x08, 0xcd, 0x08, 0xcd, 0x08, 0xcd, 0x00, 0x5c, 0x00, 0x5c, 
                         0x00, 0x5c, 0x00, 0x9c, 0x00, 0x9c, 0x00, 0x00, 0x00, 0x00, 
                         0x9c, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x19, 0x96, 0x19, 
                         0x96, 0x19, 0x96, 0x00, 0x9c, 0x00, 0x9c, 0x00, 0x00, 0x00, 
                         0x00, 0x9c, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
                         0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
                         0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x06, 
                         0x87, 0x06, 0x87, 0x06, 0x87, 0x00, 0x9c, 0x00, 0x64, 0x00, 
                         0x00, 0x00, 0x00, 0x9c])

        allcalib = AllCalibration(bin_reg)

        str_repr = str(allcalib)
        self.assertTrue('Offset bias: [0, 0, 0]' in str_repr)
        self.assertTrue('Sensitivity: [1671,' in str_repr)

